use crate::config::{Args, RaasConfig};
use crate::robot::connection_handler::ConnectionHandler;
use crate::robot::robot_manager::RobotManager;
use crate::robot::{ConnectorManager, ROBOT_MESSAGE_QUEUE_SIZE};
use log::{error, info};
use raas_types::raas::service::raas_server::RaasServer;
use server::grpc::RaaSServer;
use std::sync::Arc;
use structopt::StructOpt;
use tokio::io::duplex;
use tokio::net::TcpListener;
use tokio::sync::Mutex;
use tonic::transport::Server;

mod config;
mod robot;
mod server;

#[tokio::main]
async fn main() {
    env_logger::init();
    info!("Starting...");

    let args = Args::from_args();

    let config = RaasConfig::new(&args.cfg_path);

    let config = match config {
        Ok(c) => c,
        Err(err) => {
            error!("Error loading config: {}", err);
            return;
        }
    };

    let connector_manager = ConnectorManager {
        next_id: 0,
        connectors: Default::default(),
    };

    let connector_manager = Arc::new(Mutex::new(connector_manager));

    let tcp_listener = TcpListener::bind(&config.robot_server).await.unwrap();

    let (main_stream, manager_stream) = duplex(ROBOT_MESSAGE_QUEUE_SIZE);

    let robot_manager = RobotManager::new(manager_stream);
    let connector_handler = ConnectionHandler::new(&config.robot_server, tcp_listener);

    let robot_manager_connector_manager = connector_manager.clone();
    tokio::task::spawn(robot_manager.worker(robot_manager_connector_manager));
    tokio::task::spawn(connector_handler.worker(connector_manager));

    info!("Serving grpc at {}", &config.grpc_server);
    let raas_server = RaaSServer::new(main_stream);
    Server::builder()
        .add_service(RaasServer::new(raas_server))
        .serve(config.grpc_server.parse().unwrap())
        .await
        .unwrap()
}
