use crate::robot::{ConnectorManager, Error, ROBOT_TIMEOUT};
use log::{error, info};
use prost::Message;
use raas_types::raas;
use raas_types::raas::cmd::request::Cmd;
use raas_types::raas::cmd::Request;
use raas_types::raas::{recv_raas_msg, send_raas_msg, RaasMessage};
use std::sync::Arc;
use tokio::io::{AsyncWriteExt, DuplexStream};
use tokio::sync::Mutex;

#[derive(Debug)]
pub struct RobotManager {
    command_stream: DuplexStream,
}

impl RobotManager {
    pub fn new(stream: DuplexStream) -> Self {
        Self {
            command_stream: stream,
        }
    }

    async fn recv_command(&mut self) -> Result<RaasMessage, Error> {
        let resp = recv_raas_msg(&mut self.command_stream).await?;

        Ok(resp)
    }

    async fn send_resp(&mut self, msg: RaasMessage) -> Result<(), Error> {
        self.command_stream.write_all(&msg.into_bytes()).await?;

        Ok(())
    }

    async fn send_error(
        &mut self,
        id: u32,
        error: raas::error::ErrorType,
        err_msg: &str,
    ) -> Result<(), Error> {
        let resp = raas::resp::Response {
            id,
            timestamp: 0,
            resp: Some(raas::resp::response::Resp::Error(raas::error::Error {
                err: error.into(),
                msg: err_msg.to_string(),
            })),
        };

        let mut msg = Vec::new();

        resp.encode(&mut msg)?;

        self.send_resp(RaasMessage::new(msg)).await?;

        Ok(())
    }

    async fn handle_command(
        &mut self,
        connector_manager: Arc<Mutex<ConnectorManager>>,
    ) -> Result<(), Error> {
        info!("Waiting for command...");

        let msg = self.recv_command().await?;

        let request = Request::decode(&*msg.msg)?;

        let dest_type = match request.cmd.clone().unwrap() {
            Cmd::RollCmd(_) => raas::register::BotTypes::Roll,
        };

        let mut connector_manager = connector_manager.lock().await;
        let connector = connector_manager
            .connectors
            .iter_mut()
            .find(|(_, handler)| handler.tags.contains(&dest_type));

        if let Some((id, connector)) = connector {
            info!("Sending message to connector with id={}", id);
            let mut msg = Vec::new();
            request.encode(&mut msg)?;
            let ret = send_raas_msg(&mut connector.stream, msg).await;

            if let Err(err) = ret {
                error!(
                    "Got '{}' sending message to Robot Connector id={}, closing connection",
                    err, id
                );

                return Err(Error::ConnectionClosed(*id));
            }

            let timeout = tokio::time::timeout(ROBOT_TIMEOUT, recv_raas_msg(&mut connector.stream));

            let ret = timeout.await?;

            let resp = match ret {
                Ok(r) => r,
                Err(err) => {
                    error!(
                        "Got '{}' receiving message to Robot Connector id={}, closing connection",
                        err, id
                    );

                    return Err(Error::ConnectionClosed(*id));
                }
            };

            info!("Got response from id={}", id);
            self.send_resp(resp).await?;
        } else {
            error!("No connectors available to handle request");
            return Err(Error::NoRobotsToHandleRequest);
        }

        Ok(())
    }

    pub async fn worker(
        mut self,
        connector_manager: Arc<Mutex<ConnectorManager>>,
    ) -> Result<(), Error> {
        loop {
            let res = self.handle_command(connector_manager.clone()).await;

            if let Err(err) = res {
                error!("Got error in robot manager: {}", err);

                let (id, error, error_msg) = match err {
                    Error::ConnectionClosed(id) => {
                        let mut connection_manager = connector_manager.lock().await;

                        connection_manager.connectors.remove(&id);
                        (
                            id,
                            raas::error::ErrorType::RobotOffline,
                            "Robot offline.".to_string(),
                        )
                    }
                    Error::Io(err) => (
                        0,
                        raas::error::ErrorType::RobotError,
                        format!("IO Error: {}", err),
                    ),
                    Error::ProtobufEncode(err) => (
                        0,
                        raas::error::ErrorType::RobotError,
                        format!("Encode Error: {}", err),
                    ),
                    Error::ProtobufDecode(err) => (
                        0,
                        raas::error::ErrorType::RobotError,
                        format!("Decode Error: {}", err),
                    ),
                    Error::NoRobotsToHandleRequest => (
                        0,
                        raas::error::ErrorType::NoRobotsToHandleRequest,
                        "No robots to handle request".to_string(),
                    ),
                    Error::Timeout(_) => (
                        0,
                        raas::error::ErrorType::RobotError,
                        "Timed out waiting for response from Robot".to_string(),
                    ),
                };

                self.send_error(id, error, error_msg.as_str()).await?;
            }
        }
    }
}
